<?php

namespace InSegment\ApiCore\Api\Concerns;

use InSegment\ApiCore\Facades\DTODefs;
use Illuminate\Database\Eloquent\Collection;
use Illuminate\Database\Eloquent\Relations\BelongsTo;

use InSegment\ApiCore\Api\DataTransferObject;

trait ExportsObjects
{
    /**
     * Export one or a set of Models
     * 
     * @param \Illuminate\Database\Eloquent\Collection|array|\Illuminate\Database\Eloquent\Model $models
     * @param array|string|null $conversion
     * @param bool $useContext
     * @return $this
     */
    public function export($models, &$conversion = null, bool $useContext = false)
    {
        if (is_array($models)) {
            $models = collect($models);
        }

        $context = $useContext ? DataTransferObject::CONTEXT_EXPORT : null;
        
        if ($models instanceof Collection) {
            foreach ($models as $dotFormat => $model) {
                $this->descend(compact('model', 'dotFormat', 'context'), [$this, 'exportModel'], $conversion);
            }
        } else {
            $this->descend(['model' => $models, 'context' => $context], [$this, 'exportModel'], $conversion);
        }
        
        return $this;
    }
    
    /**
     * Export fields
     * Codes the Model data back to its original mappings
     * 
     * @param array|string|null $conversion - only this subset of base rules will apply
     * @return null
     */
    public function exportModel($conversion = null)
    {
        $ascendedOmit = $this->omit;
        foreach (array_get($this->mapping, 'omit', []) as $omit) {
            $this->omit[$omit] = true;
        }
        
        $convertedRules = $this->dtoDefs->rulesByConversion($conversion, $this->getRules($this->modelClass));
        $context = DataTransferObject::CONTEXT_EXPORT;
        
        if (!is_array($convertedRules)) {
            $this->input = $this->model->$convertedRules;
        } else {
            foreach ($convertedRules as $dotFormat => $mapping) {
                if (empty($mapping) || isset($this->omit[$dotFormat]) || isset($mapping['auto'])) {
                    continue;
                }
                
                $this->descend(compact('dotFormat', 'mapping', 'context'), [$this, 'exportMapping']);
        
                if (is_null(array_get($this->input, $dotFormat))) {
                    array_forget($this->input, $dotFormat);
                }
            }
        }
        
        if (!empty($this->context)) {
            foreach (array_keys($this->context) as $dotFormat) {
                $value = array_get($this->input, $dotFormat);
                
                if (isset($value)) {
                    $this->context[$dotFormat] = $value;
                }
            }
        }
        
        $this->omit = &$ascendedOmit;
    }
    
    /**
     * Export mapping
     * 
     * @return null
     */
    protected function exportMapping()
    {
        if (is_array($this->mapping)) {
            $typeFromMapping = array_get($this->mapping, 'type');

            if ($typeFromMapping && isset($this->mappers[$typeFromMapping])) {
                $this->mappers[$typeFromMapping]->export($this->mapping);
            }
        } else {
            $this->exportMappedField();
        }
    }
    
    /**
     * Export field mapping, supporting calling and casting, and dot notation
     * 
     * @return null
     */
    private function exportMappedField()
    {
        $target = $this->model;
        foreach (explode('.', $this->mapping) as $part) {
            $isFun = ends_with($part, '()');
            if ($isFun) {
                $part = substr($part, 0, -2);
            }
            
            $explode = explode(':', $part, 2);
            $part = $explode[0];
            $cast = $explode[1] ?? null;
            
            if (!isset($target)) {
                return;
            }
            
            if (isset($cast)) {
                if (($isFun = ends_with($part, '()'))) {
                    $part = substr($part, 0, -2);
                    $target = $target->$part();
                } else {
                    $target = data_get($target, $part);
                }
                
                switch($cast) {
                    case 'array': $target = (array) $target; break;
                    case 'string': $target = (string) $target; break;
                    case 'double': $target = (double) $target; break;
                    case 'float': $target = (float) $target; break;
                    case 'int': $target = (int) $target; break;
                }
            } else if ($isFun) {
                $target = $target->$part();
            } else {
                $target = data_get($target, $part);
            }
        }
        
        if (isset($target)) {
            $this->input = $target;
        }
    }
    
    /**
     * Export relation mapping
     * 
     * @param array $mapping
     * @return null
     */
    private function exportRelation(&$mapping)
    {
        $relation = $this->model->{$mapping['relation']}();
        $related = $this->model->{$mapping['relation']};
        
        if ($relation instanceof BelongsTo && is_string($externalKeyName = $this->externalKeyName($related, $mapping))) {
            $this->input = $related->$externalKeyName;
        } else if ($related instanceof Collection) {
            $context = DataTransferObject::CONTEXT_EXPORT;
            foreach ($related as $dotFormat => $relatedModel) {
                $this->descend(compact('dotFormat', 'context'), [$this, 'export'], $relatedModel);
            }
        } else if (isset($related)) {
            $this->export($related);
        }
    }
    
    /**
     * Get external key name by related model and mapping for its relation
     * 
     * @param \Illuminate\Database\Eloquent\Model|null $related
     * @param array $mapping
     * @return string|null
     */
    private function externalKeyName($related, &$mapping)
    {
        if (!isset($related)) {
            return null;
        }
        
        if (array_key_exists('key', $mapping)) {
            return $mapping['key'];
        }
        
        return DTODefs::getExternalKeyName(get_class($related));
    }
}
