<?php

namespace InSegment\ApiCore\Services;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Cache;

use InSegment\ApiCore\Facades\DTODefs;

class ParseCreateTable
{
    /**
     * An instance
     * 
     * @var \InSegment\ApiCore\Services\ParseCreateTable 
     */
    private static $instance = null;

    /**
     * Uuidables mentions merge
     * 
     * @var array 
     */
    private $uuidablesMentions;
    
    /**
     * An array holding show creates for tables
     * 
     * @var array 
     */
    private $showCreateData;
    
    /**
     * Prefix of service data in cache
     * 
     * @var string
     */
    private $cachePrefix;
    
    /**
     * Lists all loaded tables column unique data
     * 
     * @return array
     */
    public function &getColumnUniqueData()
    {
        return $this->showCreateData['columnUnique'];
    }
    
    /**
     * Get query to create temporary table
     * 
     * If it is insert table, AUTO_INCREMENT key will be replaced with bigint(20) suitable for UUID_SHORT()
     * CONSTRAINT's are removed from query
     * 
     * @param string $table
     * @param bool $uuidable
     * @param bool $allowInMemory
     * @return array
     */
    public function &getCreateForTemp(string $table, bool $uuidable, bool $allowInMemory)
    {
        $this->loadShowCreate($table);
        
        $uuidableTableMentions =
            ($uuidableTableMentions = array_get($this->getUuidablesMentions(), $table))
                ? array_combine($uuidableTableMentions, $uuidableTableMentions)
                : null;
        
        $showCreate = $this->showCreateData['showCreate'][$table];
        $hasMentions = false;
        $last = count($showCreate) - 1;
        
        foreach ($showCreate as $i => $part) {
            $trimmed = trim($part);
            
            if ($trimmed[0] === '`') {
                if ($uuidableTableMentions && ($keyName = $this->mentionsMatches($trimmed, $uuidableTableMentions))) {
                    $hasMentions = true;
                    $showCreate[$i] = "`{$keyName}` bigint(20) unsigned NULL,";
                } else if ($allowInMemory && preg_match('/` ((tiny|medium|long)?(text|blob)|json)[ ,\(]/i', $trimmed)) {
                    $allowInMemory = false;
                }
            } else if (starts_with($trimmed, 'CONSTRAINT')) {
                unset($showCreate[$i]);
                
                if ($last == $i + 1) {
                    for ($j = $i; $j > 0; --$j) {
                        if (isset($showCreate[$j])) {
                            $showCreate[$j] = rtrim($showCreate[$j], ',');
                            break;
                        }
                    }
                }
            }
        }
        
        if ($allowInMemory) {
            $showCreate[$last] = str_replace('ENGINE=InnoDB', 'ENGINE=MEMORY', $showCreate[$last]);
        }
        
        if ($uuidable && $hasMentions) {
            $showCreate[$last] = preg_replace('/AUTO_INCREMENT=\d+/', '', $showCreate[$last], 1);
        }

        return $showCreate;
    }
    
    /**
     * Sets the data about the primary key and unique key columns of the table
     * 
     * @param string $table
     * @param string|null $externalKeyData
     * @return $this
     */
    public function loadColumnData(string $table, $externalKeyData = null)
    {
        if (!isset($this->showCreateData['columnUnique'][$table])) {
            $this->loadShowCreate($table);
            $cacheKey = "{$this->cachePrefix}.column-unique.{$table}";
            if ($externalKeyData !== null) {
                $cacheKey .= '.ex';
            }
            
            $this->showCreateData['columnUnique'][$table] = Cache::rememberForever(
                $cacheKey, function () use ($table, $externalKeyData) {
                    if ($externalKeyData !== null) {
                        $columnUniqueForTable = [["external:{$externalKeyData}"]];
                    } else {
                        $columnUniqueForTable = [];
                    }

                    $showCreate = $this->showCreateData['showCreate'][$table];
                    $createTable = array_shift($showCreate);
                    $engineAIAndCharset = array_pop($showCreate);
                    foreach ($showCreate as $tableStatement) {
                        $tableStatement = trim($tableStatement);
                        $primary = strpos($tableStatement, 'PRIMARY KEY') === 0;
                        $unique = !$primary && strpos($tableStatement, 'UNIQUE KEY') === 0;
                        if ($primary || $unique) {
                            $matches = [];
                            if (preg_match('/\(`([^\)]+)`\)/', $tableStatement, $matches)) {
                                $columnUniqueForTable[] = explode('`,`', $matches[1]);
                            }
                        }
                    }
                    
                    return $columnUniqueForTable;
                }
            );
        }
        
        return $this;
    }
    
    /**
     * Sets the data about table creation specifics
     * 
     * @param string $table
     * @return $this
     */
    protected function loadShowCreate(string $table)
    {
        if (!isset($this->showCreateData['showCreate'][$table])) {
            $this->showCreateData['showCreate'][$table] = Cache::rememberForever(
                "{$this->cachePrefix}.show-create.{$table}", function () use ($table) {
                    return explode("\n", DB::selectOne("SHOW CREATE TABLE `{$table}`")->{'Create Table'});
                }
            );
        }
        
        return $this;
    }
    
    /**
     * Test for uuidable mentions match, remove matched from further matching
     * 
     * @param string $trimmed
     * @param array $mentions
     * @return bool|string
     */
    protected function mentionsMatches(string $trimmed, array &$mentions)
    {
        $matches = [];
        if (preg_match('/^`(' . implode('|', $mentions) . ')`/', $trimmed, $matches)) {
            unset($mentions[$matches[1]]);
            return $matches[1];
        }
        
        return false;
    }
    
    /**
     * Get merge of uuidables mention columns
     * 
     * @param array|null $allTransactionTblsToKeys
     * @return array
     */
    public function getUuidablesMentions($allTransactionTblsToKeys = null)
    {
        if (!isset($this->uuidablesMentions)) {
            $allUuidables = DTODefs::getUUIDables();
            
            if (!isset($allTransactionTblsToKeys)) {
                $allTransactionTblsToKeys = Transactor::getAllowedTransactionTablesKeys();
            }
            
            $this->uuidablesMentions = array_map(
                function ($item) {
                    return array_unique(array_wrap($item));
                }, array_merge_recursive(
                    // get base tables to keys for those tables, which are uuidable and in transaction
                    array_intersect_key($allTransactionTblsToKeys, $allUuidables),
                    // get values like ['mentioned_table' => ['some_key', ...], 'mentioned_table2' => ['some_key', ...]]
                    // from those uuidable, which base tables are in transaction
                    ...array_values(array_intersect_key($allUuidables, $allTransactionTblsToKeys))
                )
            );
        }
        
        return $this->uuidablesMentions;
    }
    
    /**
     * Get ParseCreateTable instance
     * 
     * @return \InSegment\ApiCore\Services\ParseCreateTable
     */
    public static function getInstance()
    {
        if (self::$instance == null) {
            self::$instance = new ParseCreateTable();
        }
        
        return self::$instance;
    }
    
    /**
     * A singleton, no instantiation from outside
     */
    private function __construct()
    {
        $this->showCreateData = [
            'columnUnique' => [],
            'showCreate' => []
        ];
        
        $this->cachePrefix = config('api_core.cache_prefix', 'insegment.api-core.') . 'show-create-data';
    }

}
